unit Piirto3dForm;
{------------------------------------------------------------------------------}
{
   Unit Name: Piirto3dForm
   Purpose  : sinform.pas "oliomaisemmaksi"
   Author   : Vesa Lappalainen
   Date     : 17.10.1999
   Changed  : 30.10.-99/Jani Korhonen
    + luokka cFunktionKuvaaja joka suorittaa piirtmisen.
      sinformiin lista cFunktionKuvaajista
   Changed  : 07.10.2001/vl
    + VCL/CLX compaling
   ToDo     :
}
{------------------------------------------------------------------------------}


interface

uses SysUtils, Classes,
{$ifdef CLX}
  QGraphics, QControls, QForms, QDialogs, QStdCtrls, QExtCtrls, 
{$else}
  Windows, Messages,
  Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls,
{$endif}
  m3d, QTypes ;

type
  tFunc2 = function (x,y:extended) : extended; // osoitintyyppi funktioihin
                                            // sin, exp, p2,...
  cFunktionKuvaaja2 = class
    private
      x1,x2,dx:double; // vli ja tiheys, jolla funktio piirretn
      y1,y2,dy:double; // vli ja tiheys, jolla funktio piirretn
      color   :TColor; // kuvaajan vri
      f       :tFunc2;  // osoite funktioon, joka piirretn
    public
      constructor create(ix1,ix2,iy1,iy2:double; iColor:TColor; inf:tFunc2);
      procedure Draw(dc:TCanvas; const A:cTMatrix); // kuvaajan piirto
  end;

  TFormSin = class(TForm)
    PanelButtons: TPanel;
    Button2: TButton;
    Button3: TButton;
    Button1: TButton;
    Button4: TButton;
    ButtonKierra: TButton;
    Button5: TButton;
    CBAuto: TCheckBox;
    TimerKierra: TTimer;
    PanelRot: TPanel;
    PaintBoxDraw: TPaintBox;
    procedure FormCreate(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure ButtonKierraClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure TimerKierraTimer(Sender: TObject);
    procedure CBAutoClick(Sender: TObject);
    procedure PanelRotMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure PaintBoxDrawPaint(Sender: TObject);

  private
    kiertosuunta:integer;
    bitmap : TBitmap;
    A,AOrig:cTMatrix;
    Rx,Rz,RTot : cRotMatrix;
    R:array [0..5] of cRotMatrix;
    dRect:cDRect;
    coord:cCoordinate;
//    deg: array [0..2] of double;  // kiertokulma on edelleen sinformin omaisuutta

    procedure scale;
    procedure Display;
    procedure Draw(canvas: TCanvas);
    procedure ReDraw;
    procedure kierra(i: integer);
  public

    funktiot:TList;   // Lista johon laitetaan cFunktionKuvaajat (osoitteet)
  end;

var
  FormSin: TFormSin;

implementation

{$ifdef CLX}
{-$R *.xfm}
{$else}
{$R *.dfm}
{$endif}

uses math;

//------------------------------------------------------------------------------

function msin2(x,y:extended):extended;
begin
  Result := sin(x)*cos(y);
end;

function msin(x:extended):extended;
begin
  Result := sin(x);
end;

//------------------------------------------------------------------------------
procedure TFormSin.FormCreate(Sender: TObject);
var f1:cFunktionKuvaaja2; // osoitteet cFunktionKuvaaja:iin
begin
  bitmap := TBitmap.Create;
  
  f1:=cFunktionKuvaaja2.Create(-2*PI, 2*PI,-2*PI, 2*PI,clRed,msin2);    // samat kuin malliohjel-

  funktiot:=TList.Create; // lista funktioista
  funktiot.Add(f1);       // listn funktiot (osoitteet) listaan

  A := cTMatrix.Create;
  Aorig := cTMatrix.Create;
  dRect := cDRect.Create(-2*PI,2*PI,-2*PI,2*PI);
  coord := cCoordinate.Create;
  Rtot := cRotMatrix.Create(0,0);
  Rx := cRotMatrix.Create(0,0);
  Rz := cRotMatrix.Create(0,0);
  scale;
  R[0] := cRotMatrix.Create(0,5);
  R[1] := cRotMatrix.Create(1,5);
  R[2] := cRotMatrix.Create(2,5);
  R[3] := cRotMatrix.Create(0,-5);
  R[4] := cRotMatrix.Create(1,-5);
  R[5] := cRotMatrix.Create(2,-5);
  FormResize(nil);
end;

procedure TFormSin.scale;
begin
  A.scale(dRect,GetClientRect());
  AOrig.Assign(A);
  coord.SetRect(dRect);
  A.Mul(RTot);
end;

procedure TFormSin.FormResize(Sender: TObject);
begin
  bitmap.Width := ClientWidth; bitmap.Height := ClientHeight;
  scale();
  ReDraw();
end;



procedure TFormSin.Draw(canvas:TCanvas);
var dc : TCanvas; i:integer;
    apu: cFunktionKuvaaja2;
begin
  dc := Canvas;
  coord.Draw(dc,A);
  for i:=0 to funktiot.Count-1 do begin
     apu:=funktiot.Items[i];   // osoitin := osoitin
     apu.Draw(dc,A);
  end;
end;

procedure TFormSin.Display;
begin
  PaintBoxDraw.Canvas.Draw(0,0,TGraphic(bitmap));
//  PaintBoxDraw.Canvas.CopyRect(PaintBoxDraw.ClientRect,bitmap.Canvas,PaintBoxDraw.ClientRect);
end;

procedure TFormSin.ReDraw;
begin
  bitmap.Canvas.Brush.Color := Color;
  bitmap.Canvas.Rectangle(0,0,bitmap.Width,bitmap.Height);
  Draw(bitmap.canvas);
  Display;
end;

procedure TFormSin.PaintBoxDrawPaint(Sender: TObject);
begin
  Display;
end;


procedure TFormSin.kierra(i:integer);
begin
  A.Mul(R[i]);
  RTot.Mul(R[i]);
  ReDraw;
end;

procedure TFormSin.ButtonKierraClick(Sender: TObject);
var i:integer;
begin
  i := TButton(sender).tag;
  kiertosuunta := i;
  Kierra(i);
end;

procedure TFormSin.FormDestroy(Sender: TObject);
var i:integer; apu: cFunktionKuvaaja2;
begin
  A.Free;
  dRect.Free;
  coord.Free;
  for i:=0 to 5 do R[i].Free;
  for i:=0 to funktiot.Count-1 do begin
     apu:=funktiot.Items[i];
     //funktiot.Items[i].Free;  // Miksei suora viittaus toimi???????????
     apu.Free;
  end;
  Rtot.free;
  Rx.Free;
  Rz.Free;
  funktiot.Free;
  bitmap.Free;
end;

{ cFunktionKuvaaja }
constructor cFunktionKuvaaja2.create(ix1, ix2,iy1, iy2: double; iColor: TColor; inf:TFunc2);
// konstruktorissa asetetaan rajat, vri, piirrettv funktio ja piirtovliksi
// aina (x2-x1)/50
begin
  x1:=ix1; x2:=ix2; Color:=iColor; f:=inf;
  y1:=iy1; y2:=iy2;
  dx:=(x2-x1)/50;
  dy:=(y2-y1)/50;
end;

procedure cFunktionKuvaaja2.Draw(dc: TCanvas; const A: cTMatrix);
// itse kuvaajan piirto
var y,x:double; oldcol : TColor;

  procedure x_suunta(y:double);
  var x:double;
  begin
    MoveTo(dc,A.map(x1,y,f(x1,y)));
    x := x1+dx;
    while ( x < x2 ) do begin
      LineTo(dc,A.map(x,y,f(x,y)) );
      x := x + dx;
    end;
    LineTo(dc,A.map(x2,y,f(x2,y)) );
  end;

  procedure y_suunta(x:double);
  var y:double;
  begin
    MoveTo(dc,A.map(x,y1,f(x,y1)));
    y := y1+dy;
    while ( y < y2 ) do begin
      LineTo(dc,A.map(x,y,f(x,y)) );
      y := y + dy;
    end;
    LineTo(dc,A.map(x,y2,f(x,y2)) );
  end;

begin
  oldcol := dc.Pen.Color;
  dc.Pen.Color := color;
  y := y1;
  while ( y < y2 ) do begin
    x_suunta(y);
    y := y + dy;
  end;
  x_suunta(y2);

  x := x1;
  while ( x < x2 ) do begin
    y_suunta(x);
    x := x + dx;
  end;
  y_suunta(x2);
  dc.Pen.Color := oldcol;
end;


procedure TFormSin.TimerKierraTimer(Sender: TObject);
begin
  kierra(kiertosuunta);
end;

procedure TFormSin.CBAutoClick(Sender: TObject);
begin
  TimerKierra.Enabled := CBAuto.Checked;
end;

procedure TFormSin.PanelRotMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
var degx, degz : integer; dx,dy:double;
begin
  dx := PanelRot.ClientHeight/2;         
  dy := PanelRot.ClientWidth/2;
  degz := Round((X - dx)/dx * 180);
  degx := Round((Y - dy)/dy * 180);
  A.Assign(AOrig);
  Rx.setdeg(0,degx);
  Rz.setdeg(2,degz);
  A.Mul(Rx);
  A.Mul(Rz);
  ReDraw;
end;


end.
