{------------------------------------------------------------------------------}
{
   Unit Name: m3d
   Purpose  : Hoidella 3d muunnoksia
   Author   : Vesa Lappalainen
   Date     : 17.10.1999
   Changed  : 07.10.2001/vl
    + VCL/CLX compaling

   Usage    :

}
{------------------------------------------------------------------------------}

unit m3d;

interface
uses
  Types,
{$ifdef CLX}
  QGraphics;
{$else}
  Windows,graphics;
{$endif}

type    

tFunc = function (x:extended) : extended;

//------------------------------------------------------------------------------
cDPoint = class // Reaalilukupiste
//------------------------------------------------------------------------------
private
  fx,fy : double;
public
  constructor Create(ix,iy:double) overload;
  procedure SetPoint(ix,iy:double);
  procedure Assign(pt:cDPoint);
  property x:double read fx;
  property y:double read fy;
end;

//------------------------------------------------------------------------------
cDRect = class  // Reaaliluku suorakaide;
//------------------------------------------------------------------------------
private
  fp1,fp2:cDPoint;  // Vasen alanurkka ja oikea ylnurkka
public
  constructor Create(x1:double=0;x2:double=1;y1:double=0;y2:double=1); overload;
  constructor Create(f:tFunc;x1:double=0;x2:double=0;dx:double=0); overload;
  destructor Destroy; override;
  procedure Assign(const rc:cDRect);
  procedure scan(f:tFunc;x1,x2,dx:double);
  function GetWidth():double;
  function GetHeight():double;
  function MidX():double;
  function MidY():double;
  property p1 : cDPoint read fp1;
  property p2 : cDPoint read fp2;
  property Width : double read GetWidth;
  property Height : double read GetHeight;
end;

//------------------------------------------------------------------------------
const CVSIZE=4;  // CVSIZE voi olla joko 3 (2D kuvat) tai 4 (3D kuvat)

function vsize(var i:integer):integer; // Pakotetaan aina i vlille [0,CVSIZE-1]

type
//------------------------------------------------------------------------------
cVector = class
//------------------------------------------------------------------------------
protected
private
  fa : array[0..CVSIZE-1] of double;
  function Geta(i: integer): double;
  procedure Seta(i: integer; const Value: double);
public
  constructor Create(a0:double=0.0;a1:double=0.0;a2:double=0.0); overload;
  constructor Create(ind:integer; ai:double); overload;
  procedure Assign(V:cVector);
  function InnerProduct(const b:cVector):double;
  function SetVect(ind:integer; ai:double):cVector;
  property a[i:integer]:double read Geta write Seta; default;
end;

//------------------------------------------------------------------------------
cMatrix = class
  private
    function Getr(i: integer): cVector;
    procedure Setr(i: integer; const Value: cVector);
//------------------------------------------------------------------------------
protected
  fr : array [0..CVSIZE] of cVector;
public
  constructor Create; overload;
  constructor Create(M:cMatrix); overload;
  destructor Destroy; override;
  procedure ident();
  procedure Assign(M:cMatrix);
  function column(c:integer; col : cVector): cVector;
  function Mul(const B:cMatrix) : cMatrix;
  property r[i:integer]:cVector read Getr write Setr; default;
end;

function Multiply(const A,B:cMatrix;C:cMatrix) : cMatrix; overload; // C = A*B;
function Multiply(const A:cMatrix;const x:cVector; y:cVector) : cVector; overload; // y = A*x;

type
//------------------------------------------------------------------------------
cTMatrix = class(cMatrix)   // Transformaatiomatriisi
//------------------------------------------------------------------------------
public
  constructor Create(axis:integer; deg:boolean); overload;
  procedure scale(dRect:cDRect; const iRect:TRect; equal:Boolean=true);
  function map(x:double;y:double;z:double=0):TPoint;
end;

//------------------------------------------------------------------------------
cRotMatrix = class(cMatrix)
  private
    procedure doSetDeg(axis: integer; deg: double);   // Rotaatiomatriisi
//------------------------------------------------------------------------------
public
  constructor Create(axis:integer;deg:double);
  procedure SetDeg(axis:integer;deg:double);
end;

//------------------------------------------------------------------------------
cCoordinate = class   // Piirt koordinaatiston R2:ssa
//------------------------------------------------------------------------------
  origo : cDPoint ;
  dRect : cDRect;
  ds : cDPoint;   // small = pikkutikkujen vlit
  db : cDPoint;   // big   = isojen tikkujen vlit
  Fxformat:string;
  Fyformat:string;
  procedure XTicks(dc:TCanvas; const A:cTMatrix; dx,dty:double);
  procedure YTicks(dc:TCanvas; const A:cTMatrix; dy,dtx:double);
  procedure XLabels(dc:TCanvas; const A:cTMatrix; dx,dty:double);
  procedure YLabels(dc:TCanvas; const A:cTMatrix; dy,dtx:double);
public
  constructor Create;
  destructor Destroy; override;
  procedure SetOrigo(const pt:cDPoint);
  procedure SetRect(const rc:cDRect);
  procedure SetSmallTicks(const pt:cDPoint);
  procedure SetBigTicks(const pt:cDPoint);
  procedure Draw(dc:TCanvas; const A:cTMatrix);
  property xformat : string read Fxformat write Fxformat;
  property yformat : string read Fyformat write Fyformat;
end;

//------------------------------------------------------------------------------
procedure MoveTo(dc:TCanvas;pt:TPoint);
procedure LineTo(dc:TCanvas;pt:TPoint);
procedure Draw(dc:TCanvas; const A:cTMatrix; f:tFunc;
          x1,x2,dx:double; col : TColor
          );


implementation
uses SysUtils;

{ cDPoint }

procedure cDPoint.Assign(pt: cDPoint);
begin
  SetPoint(pt.x,pt.y);
end;

constructor cDPoint.Create(ix, iy: double);
begin
  inherited Create;
  SetPoint(ix,iy);
end;


procedure cDPoint.SetPoint(ix, iy: double);
begin
  fx := ix;
  fy := iy;
end;

{ cDRect }

constructor cDRect.Create(x1, x2, y1, y2: double);
begin
  inherited Create;
  fp1 := cDPoint.Create(x1,y1);
  fp2 := cDPoint.Create(x2,y2);
end;

procedure cDRect.Assign(const rc: cDRect);
begin
  p1.Assign(rc.p1);
  p2.Assign(rc.p2);
end;

constructor cDRect.Create(f: tFunc; x1, x2, dx: double);
begin
  Create(x1,x2,0,0);
  scan(f,x1,x2,dx);
end;


destructor cDRect.Destroy;
begin
  fp1.Free;
  fp2.Free;
  inherited;
end;

function cDRect.GetHeight: double;
begin
  Result := p2.y - p1.y;
end;

function cDRect.GetWidth: double;
begin
  Result := p2.x - p1.y;
end;

function cDRect.MidX: double;
begin
  Result := ( p1.x + p2.x ) / 2.0;
end;

function cDRect.MidY: double;
begin
  Result := ( p1.y + p2.y ) / 2.0;
end;

procedure cDRect.scan(f: tFunc; x1, x2, dx: double);
var fx,x,y1,y2:double;
begin
  y1:=f(x2); y2:=f(x2);
  if ( dx = 0 ) then dx := (x2-x1)/100;
  x := x1;
  while ( x < x2 ) do begin
    fx := f(x);
    if ( fx < y1 ) then y1 := fx;
    if ( y2 < fx ) then y2 := fx;
    x := x + dx;
  end;
  if ( p1.x < x1 ) then x1 := p1.x;
  if ( p1.y < y1 ) then y1 := p1.y;
  if ( p2.x > x2 ) then x2 := p2.x;
  if ( p2.y > y2 ) then y2 := p2.y;
  p1.SetPoint(x1,y1);
  p2.SetPoint(x2,y2);
end;


function vsize(var i:integer):integer;
begin
  if ( CVSIZE <= i ) then i := CVSIZE-1;
  if ( i < 0 ) then i := 0;
  Result := i;
end;



{ cVector }

procedure cVector.Assign(V: cVector);
var i:integer;
begin
  for i:=0 to CVSIZE-1 do fa[i] := V[i];
end;

constructor cVector.Create(ind: integer; ai: double);
begin
  inherited Create;
  SetVect(ind,ai);
end;

constructor cVector.Create(a0, a1, a2: double);
var i:integer;
begin
  inherited Create;
  for i:=0 to CVSIZE-1 do fa[i] := 0.0;
  fa[0] := a0; fa[1] := a1; fa[2] := a2;
  fa[CVSIZE-1] := 1.0;
end;

function cVector.Geta(i: integer): double;
begin
  Result := fa[vsize(i)];
end;

function cVector.InnerProduct(const b: cVector): double;
var s : double; i:integer;
begin
  s := 0;
  for i:=0 to CVSIZE-1 do
    s := s + fa[i] * b.fa[i];
  Result := s;
end;

procedure cVector.Seta(i: integer; const Value: double);
begin
  fa[vsize(i)] := Value;
end;

function cVector.SetVect(ind: integer; ai: double): cVector;
var i:integer;
begin
  for i:=0 to CVSIZE-1 do fa[i] := 0.0;
  fa[vsize(ind)] := ai;
  Result := self;
end;

{ cMatrix }

procedure cMatrix.Assign(M: cMatrix);
var i:integer;
begin
  for i:=0 to CVSIZE-1 do
    fr[i].Assign(M[i]);
end;

function cMatrix.column(c: integer; col: cVector): cVector;
var i:integer;
begin
  for i:=0 to CVSIZE-1 do
    col.a[i] := r[i].a[c];
  result := col;
end;

constructor cMatrix.Create;
var i:integer;
begin
  inherited;
  for i:=0 to CVSIZE-1 do
    fr[i] := cVector.Create;
  ident;
end;

constructor cMatrix.Create(M: cMatrix);
begin
  Create;
  Assign(M);
end;

destructor cMatrix.Destroy;
var i:integer;
begin
  for i:=0 to CVSIZE-1 do
    fr[i].Free;
  inherited;
end;

function cMatrix.Getr(i: integer): cVector;
begin
  Result := fr[vsize(i)];
end;

procedure cMatrix.ident;
var i:integer;
begin
  for i:=0 to CVSIZE-1 do
    r[i].SetVect(i,1);
end;

function cMatrix.Mul(const B: cMatrix): cMatrix;
var C : CMatrix;
begin
  C := cMatrix.Create(self);
  Multiply(C,B,self);
  Result := self;
end;

procedure cMatrix.Setr(i: integer; const Value: cVector);
begin
  fr[i].Assign(Value);
end;

function Multiply(const A,B:cMatrix;C:cMatrix) : cMatrix; // C = A*B;
var i,j : integer; col : cVector;
begin
  col := cVector.Create;
  for j:=0 to CVSIZE-1 do
    for i:=0 to CVSIZE-1 do
      C[j][i] := A[j].InnerProduct(B.column(i,col));
  Result := C;
  col.Free;
end;

function Multiply(const A:cMatrix;const x:cVector; y:cVector) : cVector; // y = A*x;
var j:integer;
begin
  for j:=0 to CVSIZE-1 do
    y[j] := A[j].InnerProduct(x);
  Result := y;
end;

{ cTMatrix }

constructor cTMatrix.Create(axis: integer; deg: boolean);
begin
  inherited Create;
end;

function cTMatrix.map(x, y, z: double): TPoint;
// Kutsulla A(x,y) muutetaan reaalimaailman piste nytn pisteeksi
var vr,vs:cVector; p:TPoint;
begin
  vr := cVector.Create(x,y,z);        // Reaalimaailman vektori
  vs := cVector.Create;
  Multiply(self, vr, vs );                 // Nytll vastaava vektori
  p.x := Round(vs[0]); p.y := Round(vs[1]);
  Result := p;
  vr.Free; vs.Free;
end;

procedure cTMatrix.scale(dRect: cDRect; const iRect: TRect; equal: Boolean);
var  scale,xs,ys,dxf,dyf,dxt,dyt:double; // s=scale, f=from,  t=to
begin
  ident;
  dxf := dRect.Width;
  dyf := dRect.Height;
  dxt := iRect.right - iRect.left;
  dyt := iRect.top - iRect.bottom;

  if ( dxf = 0 ) or ( dxt = 0 ) then xs := 1 else xs := dxt/dxf;
  if ( dyf = 0 ) or ( dyt = 0 ) then ys := 1 else ys := dyt/dyf;

  if ( equal ) then begin // Jos halutaan x/y = 1
    scale := abs(xs); if ( abs(ys) < scale ) then scale := abs(ys);
    xs := abs(xs)/xs*scale;  ys := abs(ys)/ys*scale;
  end;

  self[0][0] := xs;  self[1][1] := ys;

  self[0][CVSIZE-1] := (iRect.left+iRect.right)/2.0 - dRect.MidX()*xs;
  self[1][CVSIZE-1] := (iRect.top+iRect.bottom)/2.0 - dRect.MidY()*ys;
end;

{ cRotMatrix }
procedure cRotMatrix.doSetDeg(axis: integer; deg: double);
// Kierto akselin ympri (0=x,1=y,2=z)
//   0:   1  0  0    1:   c  0 -s    2:   c -s  0
//        0  c -s         0  1  0         s  c  0
//        0  s  c         s  0  c         0  0  1
var a,ca,sa:double;
    j : array[0..2] of integer; i,ji:integer;
begin
  inherited Create;
  a := deg*PI/180.0; ca := cos(a); sa :=sin(a);
  ji := 0;
  for i:=0 to 3-1 do  if ( i <> axis ) then begin j[ji] := i; inc(ji); end;
  r[j[0]][j[0]] :=  ca;
  r[j[1]][j[1]] :=  ca;
  r[j[0]][j[1]] := -sa;
  r[j[1]][j[0]] := +sa;
end;

constructor cRotMatrix.Create(axis: integer; deg: double);
begin
  doSetDeg(axis,deg);
end;

procedure cRotMatrix.SetDeg(axis: integer; deg: double);
begin
  ident();
  doSetDeg(axis,deg);
end;

{ cCoordinate }

constructor cCoordinate.Create;
begin
  inherited;
  origo := cDPoint.Create(0,0);
  dRect := cDRect.Create;
  ds := cDPoint.Create(0.1,0.1);   // small = pikkutikkujen vlit
  db := cDPoint.Create(1.0,1.0);   // big   = isojen tikkujen vlit
  Fxformat := '%3.1f';
  Fyformat := '%3.1f';

end;

destructor cCoordinate.Destroy;
begin
  origo.Free;
  dRect.Free;
  ds.Free;
  db.Free;
  inherited;

end;

procedure MoveTo(dc:TCanvas;pt:TPoint);
begin
  dc.MoveTo(pt.x,pt.y);
end;

procedure LineTo(dc:TCanvas;pt:TPoint);
begin
  dc.LineTo(pt.x,pt.y);
end;

procedure TextOut(dc:TCanvas;pt,dp:TPoint;const s:string);
begin
  pt.x := pt.x + dp.x;
  pt.y := pt.y + dp.y;
  dc.TextOut(pt.x,pt.y,s);
end;

procedure cCoordinate.Draw(dc: TCanvas; const A: cTMatrix);
begin
  MoveTo(dc,A.map(dRect.p1.x,origo.y));
  LineTo(dc,A.map(dRect.p2.x,origo.y));

  XTicks(dc,A,ds.x,dRect.Height/100);
  XTicks(dc,A,db.x,dRect.Height/40);
  XLabels(dc,A,db.x,dRect.Height/40);

  MoveTo(dc,A.map(origo.x,dRect.p1.y));
  LineTo(dc,A.map(origo.x,dRect.p2.y));

  YTicks(dc,A,ds.x,dRect.Width/100);
  YTicks(dc,A,db.x,dRect.Width/40);
  YLabels(dc,A,db.x,dRect.Width/30);

end;

procedure cCoordinate.SetBigTicks(const pt: cDPoint);
begin
  db.Assign(pt);
end;

procedure cCoordinate.SetOrigo(const pt: cDPoint);
begin
  origo.Assign(pt);
end;

procedure cCoordinate.SetRect(const rc: cDRect);
begin
  dRect.Assign(rc);
end;

procedure cCoordinate.SetSmallTicks(const pt: cDPoint);
begin
  ds.Assign(pt);
end;

procedure cCoordinate.XLabels(dc: TCanvas; const A: cTMatrix; dx,
  dty: double);
var x,y1:double; s:string;  pt : TPoint;
    oldStyle : TBrushStyle;
begin
  oldStyle := dc.Brush.Style;
  dc.Brush.Style := bsClear;

  y1  := origo.y-dty;
  x := origo.x-dx;
  pt.y := 0;

  while ( x >= dRect.p1.x ) do begin
    s := Format(xformat,[x]);
    pt.x := - dc.TextWidth(s) div 2;
    TextOut(dc,A.map(x,y1),pt,s);
    x := x - dx;
  end;
  x := origo.x+dx;
  while (  x <= dRect.p2.x ) do begin
    s := Format(xformat,[x]);
    pt.x := - dc.TextWidth(s) div 2;
    TextOut(dc,A.map(x,y1),pt,s);
    x := x + dx;
  end;
  dc.Brush.Style := oldStyle;
end;

procedure cCoordinate.XTicks(dc: TCanvas; const A: cTMatrix; dx,
  dty: double);
var x,y1,y2 : double;
begin
  y1  := origo.y-dty;
  y2  := origo.y+dty;
  x := origo.x;
  while ( x >= dRect.p1.x ) do begin
    MoveTo(dc,A.map(x,y1)); LineTo(dc,A.map(x,y2));
    x := x - dx;
  end;
  x := origo.x;
  while ( x <= dRect.p2.x ) do begin
    MoveTo(dc,A.map(x,y1)); LineTo(dc,A.map(x,y2));
    x := x + dx;
  end;

end;

procedure cCoordinate.YLabels(dc: TCanvas; const A: cTMatrix; dy,
  dtx: double);
var y,x1:double; s:string; pt : TPoint;
    oldStyle : TBrushStyle;
begin
  oldStyle := dc.Brush.Style;
  dc.Brush.Style := bsClear;

  pt.y := -dc.TextHeight('X') div 2;
  x1 := origo.x-dtx;
  y := origo.y-dy;
  while ( y >= dRect.p1.y ) do begin
    s := Format(yformat,[y]);
    pt.x := -dc.TextWidth(s);
    TextOut(dc,A.map(x1,y),pt,s);
    y := y - dy;
  end;
  y := origo.y+dy;
  while ( y <= dRect.p2.y ) do begin
    s := Format(yformat,[y]);
    pt.x := -dc.TextWidth(s);
    TextOut(dc,A.map(x1,y),pt,s);
    y := y + dy;
  end;
  dc.Brush.Style := oldStyle;
end;

procedure cCoordinate.YTicks(dc: TCanvas; const A: cTMatrix; dy,
  dtx: double);
var y,x1,x2 : double;
begin
  x1  := origo.x-dtx;
  x2  := origo.x+dtx;
  y := origo.y;
  while ( y >= dRect.p1.y ) do begin
    MoveTo(dc,A.map(x1,y)); LineTo(dc,A.map(x2,y));
    y := y - dy;
  end;
  y := origo.y;
  while ( y <= dRect.p2.y ) do begin
    MoveTo(dc,A.map(x1,y)); LineTo(dc,A.map(x2,y));
    y := y + dy;
  end;

end;

//------------------------------------------------------------------------------
procedure Draw(dc:TCanvas; const A:cTMatrix; f:tFunc;
          x1,x2,dx:double; col : TColor
          );
//------------------------------------------------------------------------------
var x:double; oldcol : TColor;
begin
  oldcol := dc.Pen.Color;
  dc.Pen.Color := col;
  MoveTo(dc,A.map(x1,f(x1)));
  x := x1+dx;
  while ( x < x2 ) do begin
    LineTo(dc,A.map(x,f(x)));
    x := x + dx;
  end;
  LineTo(dc,A.map(x2,f(x2)));
  dc.Pen.Color := oldcol;
end;


end.
